# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import time
import pytest
import pandas as pd
from mitosheet.streamlit.v1.spreadsheet import get_dataframe_hash


TEST_HASHES = [
    (
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}),
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}),
        True
    ),
    # Extra column
    (
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}),
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4], 'C': [1, 2, 3]}),
        False
    ),
    # Different values
    (
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}),
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 5]}),
        False
    ),
    # Different order
    (
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}),
        pd.DataFrame(data={'B': [1, 2, 3], 'A': [2, 3, 4]}),
        False
    ),
    # Different index
    (
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}, index=[1, 2, 3]),
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}, index=[1, 2, 4]),
        False
    ),
    # Missing column
    (
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}),
        pd.DataFrame(data={'A': [1, 2, 3]}),
        False
    ),
    # Different sort
    (
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}).sort_values(by=['A']),
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}).sort_values(by=['B'], ascending=False),
        False
    ),
    # Different headers
    (
        pd.DataFrame(data={'A': [1, 2, 3], 'C': [2, 3, 4]}),
        pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}),
        False
    ),
    # Reordered columns
    (
        pd.DataFrame(data={'A': [1, 2, 3], 'C': [2, 3, 4]}),
        pd.DataFrame(data={'C': [2, 3, 4], 'A': [1, 2, 3]}),
        False
    ),
]

@pytest.mark.parametrize("df1, df2, expected", TEST_HASHES)
def test_hash_pandas_dataframe(df1, df2, expected):
    assert len(get_dataframe_hash(df1)) < 100
    assert (get_dataframe_hash(df1) == get_dataframe_hash(df2)) == expected
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.pivot_code_chunk import \
    PivotCodeChunk
from mitosheet.errors import make_no_column_error
from mitosheet.state import DATAFRAME_SOURCE_PIVOTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import StepType

# Aggregation types pivot supports
PA_COUNT_UNIQUE = 'count unique'
PIVOT_AGGREGATION_TYPES = [
    # These first few are supported out of the box by 
    # pandas, so we don't need any extra support for them
    'sum',
    'mean',
    'median',
    'min',
    'max', 
    'count', 
    'std',
    PA_COUNT_UNIQUE
]

# Pivot Column Transformations: if the user does not specify a transformation before the pivot (e.g. 
# getting the months from a date), then we call this a no-op
PCT_NO_OP = 'no-op'
PCT_DATE_YEAR = 'year'
PCT_DATE_QUARTER = 'quarter'
PCT_DATE_MONTH = 'month'
PCT_DATE_WEEK = 'week'
PCT_DATE_DAY_OF_MONTH = 'day of month'
PCT_DATE_DAY_OF_WEEK = 'day of week'
PCT_DATE_HOUR = 'hour'
PCT_DATE_MINUTE = 'minute'
PCT_DATE_SECOND = 'second'
PCT_DATE_YEAR_MONTH_DAY_HOUR_MINUTE = 'year-month-day-hour-minute'
PCT_DATE_YEAR_MONTH_DAY_HOUR = 'year-month-day-hour'
PCT_DATE_YEAR_MONTH_DAY = 'year-month-day'
PCT_DATE_YEAR_MONTH = 'year-month'
PCT_DATE_YEAR_QUARTER = 'year-quarter'
PCT_DATE_MONTH_DAY = 'month-day'
PCT_DATE_DAY_HOUR = 'day-hour'
PCT_DATE_HOUR_MINUTE = 'hour-minute'


class PivotStepPerformer(StepPerformer):
    """
    A pivot, which allows you to pivot data from an existing dataframe 
    along some keys, and then aggregate columns with specific functions.
    """

    @classmethod
    def step_version(cls) -> int:
        return 9

    @classmethod
    def step_type(cls) -> str:
        return 'pivot'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any], previous_steps: List[StepType]) -> Dict[str, Any]:
        """
        We filter out any duplicated aggregation keys, as they
        result in errors without adding any data to the pivot.
        """
        # Filter out any duplicate aggregation functions
        for column_id, aggregation_function_names in params['values_column_ids_map'].items():
            new_aggregation_function_names = []
            for i in aggregation_function_names:
                if i not in new_aggregation_function_names:
                    new_aggregation_function_names.append(i)
            params['values_column_ids_map'][column_id] = new_aggregation_function_names

        # Add in any optional code that we want to try and execute
        # which in this case are the edits we replay on top of the 
        # pivot table
        optional_code, optional_code_chunk_names = get_optional_code_to_replay_on_dataframe_creation(previous_steps, params)
        params['optional_code'] = optional_code
        params['optional_code_chunk_names'] = optional_code_chunk_names

        return params
    

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        destination_sheet_index: Optional[int] = get_param(params, 'destination_sheet_index')
        
        final_destination_sheet_index = destination_sheet_index if destination_sheet_index is not None else len(prev_state.dfs)

        new_df_name = get_new_pivot_df_name(prev_state, sheet_index) if destination_sheet_index is None else prev_state.df_names[destination_sheet_index]

        execution_data = {
            # NOTE: we return the final destination name so frontend pivots can match on this execution data - it's not
            # actually used by the backend (which is kinda confusing and should be renamed)
            'destination_sheet_index': final_destination_sheet_index,
            'new_df_name': new_df_name,
        }

        try:
            return cls.execute_through_transpile(
                prev_state, 
                params, 
                execution_data,
                new_dataframe_params={
                    'df_source': DATAFRAME_SOURCE_PIVOTED,
                    'new_df_names': [new_df_name],
                    'overwrite': {
                        'sheet_index_to_overwrite': destination_sheet_index,
                        'attempt_to_save_filter_metadata': True
                    } if destination_sheet_index is not None else destination_sheet_index
                },
                optional_code=params.get('optional_code'),
                use_deprecated_id_algorithm=get_param(params, 'use_deprecated_id_algorithm')
            )
        except KeyError as e:
            column_header = e.args[0]
            raise make_no_column_error([column_header], error_modal=False)


    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            PivotCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'destination_sheet_index'),
                get_param(params, 'pivot_rows_column_ids_with_transforms'),
                get_param(params, 'pivot_columns_column_ids_with_transforms'),
                get_param(params, 'pivot_filters'),
                get_param(params, 'values_column_ids_map'),
                get_param(params, 'flatten_column_headers'),
                get_param(params, 'public_interface_version'),
                get_param(execution_data if execution_data is not None else {}, 'new_df_name'),
                get_param(execution_data if execution_data is not None else {}, 'optional_code_that_successfully_executed'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        destination_sheet_index = get_param(params, 'destination_sheet_index')
        if destination_sheet_index is not None: # If editing an existing sheet, that is what is changed
            return {destination_sheet_index}
        return {-1}
    

def get_new_pivot_df_name(prev_state: State, sheet_index: int) -> str: 
    """
    Creates the name for the new pivot table sheet using the format
    {source_sheet_name}_{pivot} or {source_sheet_name}_{pivot}_1, etc. 
    if the pivot table name already exists. 
    """
    new_df_name_original = prev_state.df_names[sheet_index] + '_pivot'
    curr_df_name = new_df_name_original
    multiple_sheet_indicator = 1
    while curr_df_name in prev_state.df_names:
        curr_df_name = f'{new_df_name_original}_{multiple_sheet_indicator}'
        multiple_sheet_indicator += 1
    return curr_df_name


def get_optional_code_to_replay_on_dataframe_creation(previous_steps: List[StepType], params: Dict[str, Any]) -> Tuple[Tuple[List[str], List[str]], List[str]]:
    """
    This function is a utility used to replay edits on top of pivot tables (or merges) - specifically, 
    it finds the code to try and execute on top of the a dataframe that is being edited.    
    """
    # If it does not have a destination sheet index, then we're going to 
    # go look for the original step that created this dataframe
    if params.get('destination_sheet_index') is None:
        return ([], []), []

    destination_sheet_index = params['destination_sheet_index']

    optional_code: Tuple[List[str], List[str]] = ([], [])

    # First, find the index of the step we have to start looking from
    # for steps that edit the pivot table
    starting_index = None
    for index, step in enumerate(reversed(previous_steps)):
        # Case 1: we find the last step that edited this pivot, and get it's extra code 
        # that successfully executed, as we want to start by replaying this
        if (step.step_type == 'pivot' or step.step_type == 'merge') and 'destination_sheet_index' in step.params and step.params['destination_sheet_index'] == destination_sheet_index:
            optional_code = step.execution_data.get('optional_code_that_successfully_executed', ([], []))
            starting_index = len(previous_steps) - index - 1
            break

        # Case 2: we find the step that created this extra dataframe
        elif (step.step_type == 'pivot' or step.step_type == 'merge') and len(step.dfs) == destination_sheet_index + 1:
            starting_index = len(previous_steps) - index - 1
            break
    
    # Then, go through and find all of the steps after this creating step that 
    # modify the pivot table only -- and we get the code for all of them
    code_chunk_names = []
    if starting_index is not None:
        previous_steps = previous_steps[starting_index + 1:]
        previous_steps = [step for step in previous_steps if step.step_performer.get_modified_dataframe_indexes(step.params) == {destination_sheet_index}]
        code_chunks: List["CodeChunk"] = []
        for step in previous_steps:
            code_chunks += step.step_performer.transpile(step.initial_defined_state, step.params, step.execution_data)
        
        all_import_code = []
        all_other_code = []
        for code_chunk in code_chunks:
            comment = '# ' + code_chunk.get_description_comment().strip().replace('\n', '\n# ')
            code_lines, import_lines = code_chunk.get_code()
            all_import_code += import_lines
            all_other_code += ['', comment]
            all_other_code += code_lines
            code_chunk_names.append(type(code_chunk).__name__)

        optional_code = (
            optional_code[0] + all_other_code,
            optional_code[1] + all_import_code,
        )

    return optional_code, code_chunk_names


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from packaging.version import Version
from typing import Any, Dict, List
import warnings
import pytest
import pandas as pd

from mitosheet.errors import MitoError
from mitosheet.parser import get_backend_formula_from_frontend_formula, parse_formula, safe_contains, get_frontend_formula
from mitosheet.types import FORMULA_ENTIRE_COLUMN_TYPE, FORMULA_SPECIFIC_INDEX_LABELS_TYPE
from mitosheet.tests.decorators import pandas_post_1_2_only


def get_number_data_for_df(columns: List[Any], length: int) -> Dict[Any, Any]:
    return {ch: [i for i in range(length)] for ch in columns}


def get_string_data_for_df(columns: List[Any], length: int) -> Dict[Any, Any]:
    return {ch: [str(i) for i in range(length)] for ch in columns}

CONSTANT_TEST_CASES: Any = [
    (
        '=100',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = 100',
        set([]),
        set([])
    ),
    (
        '    =100',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = 100',
        set([]),
        set([])
    ),
    (
        '\t\n=100',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = 100',
        set([]),
        set([])
    ),
    (
        '=100\n+100',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = 100+100',
        set([]),
        set([])
    ),
    (
        '=100\t+100',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = 100+100',
        set([]),
        set([])
    ),
    (
        '=True',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = True',
        set([]),
        set([])
    ),
    (
        '=TRUE',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = True',
        set([]),
        set([])
    ),
    (
        '=true',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = True',
        set([]),
        set([])
    ),
    (
        '=true0',
        'true',
        0,
        pd.DataFrame(get_string_data_for_df(['true'], 2)),
        'df[\'true\'] = df[\'true\']',
        set([]),
        {'true'}
    ),
    (
        '=TRUE0',
        'TRUE',
        0,
        pd.DataFrame(get_string_data_for_df(['TRUE'], 2)),
        'df[\'TRUE\'] = df[\'TRUE\']',
        set([]),
        {'TRUE'}
    ),
    (
        '=\'StringSingleQuotes\'',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \'StringSingleQuotes\'',
        set([]),
        set([])
    ),
    (
        '=\"StringDoubleQuotes\"',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \"StringDoubleQuotes\"',
        set([]),
        set([])
    ), 
    (
        '=\"String Double Quotes\"',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \"String Double Quotes\"',
        set([]),
        set([])
    ), 
    (
        '=FUNC(\"String Double Quotes\")',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = FUNC(\"String Double Quotes\")',
        set(['FUNC']),
        set([])
    ),
    (
        '=FUNC(\"DIFF_FUNC(A)\")',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = FUNC(\"DIFF_FUNC(A)\")',
        set(['FUNC']),
        set([])
    ),
    (
        '=\"String One FUNC(A)\" + \"STRING TWO FUNC(A)\"',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \"String One FUNC(A)\" + \"STRING TWO FUNC(A)\"',
        set([]),
        set([])
    ),
    (
        '=\'String One FUNC(A)\' + \'STRING TWO FUNC(A)\'',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \'String One FUNC(A)\' + \'STRING TWO FUNC(A)\'',
        set([]),
        set([])
    ),
    (
        '=FUNC1(\'String One FUNC(A)\') + FUNC2(\'STRING TWO FUNC(A)\')',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = FUNC1(\'String One FUNC(A)\') + FUNC2(\'STRING TWO FUNC(A)\')',
        set(['FUNC1', 'FUNC2']),
        set([])
    ),
    (
        '=A\nA + 100',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B', 'A\nA'], 2)),
        'df[\'B\'] = df[\'A\\nA\'] + 100',
        set([]),
        set(['A\nA'])
    ),
    (
        '=A\tA + 100',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B', 'A\tA'], 2)),
        'df[\'B\'] = df[\'A\\tA\'] + 100',
        set([]),
        set(['A\tA'])
    ),
    (
        '=CONCAT(A, "=123")',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B', 'A'], 2)),
        'df[\'B\'] = CONCAT(df[\'A\'], "=123")',
        set(['CONCAT']),
        set(['A'])
    ),
    (
        '=" = "',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = " = "',
        set([]),
        set([])
    ),
    (
        '="="',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = "="',
        set([]),
        set([])
    ),
    (
        '="=="',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = "=="',
        set([]),
        set([])
    ),
    (
        "'=='",
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        "df[\'B\'] = '=='",
        set([]),
        set([])
    ),
]

# Tests cases to ensure operators are parsed correctly
OPERATOR_TEST_CASES = [
    # Simple add operation
    (
        '=A + B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A\'] + df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Simple subtract operation
    (
        '=A - B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A\'] - df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Simple multiply operation
    (
        '=A * B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A\'] * df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Simple divide operation
    (
        '=A / B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A\'] / df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Handles multi-char columns
    (
        '=AAA + BBB + CCC + DDD',
        'E',
        0,
        pd.DataFrame(get_number_data_for_df(['AAA', 'BBB', 'CCC', 'DDD'], 2)),
        'df[\'E\'] = df[\'AAA\'] + df[\'BBB\'] + df[\'CCC\'] + df[\'DDD\']',
        set(),
        set(['AAA', 'BBB', 'CCC', 'DDD'])
    ),
    # Maintains parens and constants
    (
        '=(A + B) / C + A * 100',
        'D',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'D\'] = (df[\'A\'] + df[\'B\']) / df[\'C\'] + df[\'A\'] * 100',
        set(),
        set(['A', 'B', 'C'])
    ),
    # Operator in functions
    (
        '=FUNC(A + B / C + A * 100)',
        'D',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'D\'] = FUNC(df[\'A\'] + df[\'B\'] / df[\'C\'] + df[\'A\'] * 100)',
        set(['FUNC']),
        set(['A', 'B', 'C'])
    ),
    # Equality Operator
    (
        '=A == B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A\'] == df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Equality Operator with Equals Sign in Column Header
    (
        '=A=123 == B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A=123', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A=123\'] == df[\'B\']',
        set(),
        set(['A=123', 'B'])
    ),
    # If equals sign is in a column header, then we 
    # don't check for single equals sings because its too hard and rare
    (
        '=A=123 = B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A=123', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A=123\'] = df[\'B\']',
        set(),
        set(['A=123', 'B'])
    ),
]

# Tests proper function parsing
FUNCTION_TEST_CASES = [
    # Simple, singular function call
    (
        '=FUNC(A)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'B\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Simple, singular function call that has the name of a column
    (
        '=C(A)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'B\'] = C(df[\'A\'])',
        set(['C']),
        set(['A'])
    ),
    # Nested functions with different names
    (
        '=FUNCA(FUNCB(FUNCC(FUNCD(FUNCE(FUNCF(FUNCG(FUNCH(A))))))))',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'B\'] = FUNCA(FUNCB(FUNCC(FUNCD(FUNCE(FUNCF(FUNCG(FUNCH(df[\'A\']))))))))',
        set(['FUNCA', 'FUNCB', 'FUNCC', 'FUNCD', 'FUNCE', 'FUNCF', 'FUNCG', 'FUNCH']),
        set(['A'])
    ),
    # Nested functions with basic operators
    (
        '=RIGHT(A, LEN(B) - 10)',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = RIGHT(df[\'A\'], LEN(df[\'B\']) - 10)',
        set(['RIGHT', 'LEN']),
        set(['A', 'B'])
    ),
    # More nested functions with columns
    (
        '=FUNC(A, FUNC(B, FUNC(C)))',
        'D',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'D\'] = FUNC(df[\'A\'], FUNC(df[\'B\'], FUNC(df[\'C\'])))',
        set(['FUNC']),
        set(['A', 'B', 'C'])
    ),
    # Lots of column references
    (
        '=FUNC(A, B, C, D, E, F, G, H, I)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I'], 2)),
        'df[\'J\'] = FUNC(df[\'A\'], df[\'B\'], df[\'C\'], df[\'D\'], df[\'E\'], df[\'F\'], df[\'G\'], df[\'H\'], df[\'I\'])',
        set(['FUNC']),
        set(['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I'])
    ),
    # Lots of column references to columns with numbers
    (
        '=FUNC(column1, column2)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['column1', 'column2', 'J'], 2)),
        'df[\'J\'] = FUNC(df[\'column1\'], df[\'column2\'])',
        set(['FUNC']),
        set(['column1', 'column2'])
    ),
    # Test references to columns with spaces
    (
        '=FUNC(column 1, column 2)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['column 1', 'column 2', 'J'], 2)),
        'df[\'J\'] = FUNC(df[\'column 1\'], df[\'column 2\'])',
        set(['FUNC']),
        set(['column 1', 'column 2'])
    ),
    # Test references to columns that are numbers
    (
        '=FUNC(123, 456)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df([123, 456, 'J'], 2)),
        'df[\'J\'] = FUNC(df[123], df[456])',
        set(['FUNC']),
        set([123, 456])
    ),
    # Test references to columns that are numbers and are also strings
    (
        '=FUNC(123, 456, abc)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['abc', 456, 123, "J"], 2)),
        'df[\'J\'] = FUNC(df[123], df[456], df[\'abc\'])',
        set(['FUNC']),
        set([123, 456, 'abc'])
    ),
    # Test references to strings that has a single quotes in it
    (
        '=FUNC(this is a string \' that has a quote, 456, abc)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['abc', 456, 'this is a string \' that has a quote', "J"], 2)),
        'df[\'J\'] = FUNC(df["this is a string \' that has a quote"], df[456], df[\'abc\'])',
        set(['FUNC']),
        set(['abc', 456, 'this is a string \' that has a quote'])
    ),
    # Test references to strings that have mulitple quotes
    (
        '=FUNC(this is a string \'test\' that has a quote, 456, abc)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['abc', 456, 'this is a string \'test\' that has a quote', "J"], 2)),
        'df[\'J\'] = FUNC(df["this is a string \'test\' that has a quote"], df[456], df[\'abc\'])',
        set(['FUNC']),
        set(['abc', 456, 'this is a string \'test\' that has a quote'])
    ),
    # Test references to a string in quotes, but is a column header
    (
        '=FUNC("column")',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(["\"column\"", "J"], 2)),
        'df[\'J\'] = FUNC(df[\'"column"\'])',
        set(['FUNC']),
        set(["\"column\""])
    ),
    # Test references to a quote that is covered by a column header
    (
        '=FUNC("this is a string")',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(["string", "J"], 2)),
        'df[\'J\'] = FUNC("this is a string")',
        set(['FUNC']),
        set([])
    ),
    # Test references a boolean column
    (
        '=true + false',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df([True, False], 2)),
        'df[\'A\'] = df[True] + df[False]',
        set([]),
        set([True, False])
    ),
    # Test references a multi-index header
    (
        '=Height, min',
        'Name',
        0,
        pd.DataFrame(get_number_data_for_df([('Name', ''), ('Height', 'min')], 2)),
        'df[\'Name\'] = df[(\'Height\', \'min\')]',
        set([]),
        set([('Height', 'min')])
    ),
    # Test references a multi-index header in a function
    (
        '=FUNC(Height, min)',
        'Name',
        0,
        pd.DataFrame(get_number_data_for_df([('Name', ''), ('Height', 'min')], 2)),
        'df[\'Name\'] = FUNC(df[(\'Height\', \'min\')])',
        set(['FUNC']),
        set([('Height', 'min')])
    ),
    # Test references a multi-index header with an empty second item
    (
        '=FUNC(Name)',
        'Name',
        0,
        pd.DataFrame(get_number_data_for_df([('Name', ''), ('Height', 'min')], 2)),
        'df[\'Name\'] = FUNC(df[(\'Name\', \'\')])',
        set(['FUNC']),
        set([('Name', '')])
    ),
    # Test references a header with characters that make an invalid regex
    (
        '=FUNC(C++)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'C++'], 2)),
        'df[\'A\'] = FUNC(df[\'C++\'])',
        set(['FUNC']),
        set(['C++'])
    ),
]

# To supress warnings, we create these variables here
with warnings.catch_warnings():
    # Check that the pandas verison is < 2.0
    if Version(pd.__version__) < Version('2.0'):
        warnings.simplefilter('ignore')
        unit64_index = pd.UInt64Index(range(10))
        float64_index = pd.Float64Index(range(10))
    else:
        unit64_index = pd.Index(range(10), dtype='uint64')
        float64_index = pd.Index(range(10), dtype='float64')


INDEX_TEST_CASES = [
    # Test references a range index with a valid number 0
    (
        'A0',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = df[\'A\']',
        set([]),
        set(['A'])
    ),
    # Test references a range index with a valid number 0, with two
    (
        'A0 + A0',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = df[\'A\'] + df[\'A\']',
        set([]),
        set(['A'])
    ),
    # Test references a range index with a valid number 0, function
    (
        '=FUNC(A0)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index with a valid number non 0
    (
        '=FUNC(A0, A1)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index with a valid number non 0, positive shift
    (
        '=FUNC(A0, A1)',
        'A',
        1,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'].shift(1, fill_value=0), df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index with a number that is NOT part of the index
    (
        '=FUNC(A0, A2)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'], A2)',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index that does not start at 0 for proper offset
    (
        '=FUNC(A10, A11)',
        'A',
        10,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(10, 12)),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index that does not start at 0 for proper offset, positive shift
    (
        '=FUNC(A10, A11)',
        'A',
        11,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(10, 12)),
        'df[\'A\'] = FUNC(df[\'A\'].shift(1, fill_value=0), df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a different number index
    (
        '=FUNC(A0, A2)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 10), index=unit64_index),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references an index with a number dtype
    (
        '=FUNC(A0.0, A2.2)',
        'A',
        0.0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.Index([0.0, 1.1, 2.2], dtype='float64')),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references an index with a number dtype, positive shift
    (
        '=FUNC(A0.0, A2.2)',
        'A',
        2.2,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.Index([0.0, 1.1, 2.2], dtype='float64')),
        'df[\'A\'] = FUNC(df[\'A\'].shift(2, fill_value=0), df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index not starting from 0
    (
        '=FUNC(A1)',
        'A',
        1,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 1), index=pd.RangeIndex(1, 2)),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index not starting from below 0
    (
        '=FUNC(A-2)',
        'A',
        -2,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 4), index=pd.RangeIndex(-2, 2)),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index not starting from below 0, positive shift
    (
        '=FUNC(A1, A-2)',
        'A',
        1,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 4), index=pd.RangeIndex(-2, 2)),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(3, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index starting with a decimal
    (
        '=FUNC(A0.1, A0.2)',
        'A',
        .1,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.Index([.1, .2, .3])),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference no offset
    (
        '=FUNC(A2007-01-22 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 1), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference with offset of 1 day
    (
        '=FUNC(A2007-01-23 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference with offset of 1 day, positive shift
    (
        '=FUNC(A2007-01-22 00:00:00)',
        'A',
        pd.to_datetime('2007-01-23 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference with offset of 1 month, should still only have offset of 1
    (
        '=FUNC(A2007-02-22 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-02-22 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference with no offset, and some offset
    (
        '=FUNC(A2007-02-22 00:00:00, A2007-01-22 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-02-22 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0), df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index with no offset
    (
        '=FUNC(Aa)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 1), index=pd.Index(['a'])),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index with at the end of the formula
    (
        '=Aa',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 1), index=pd.Index(['a'])),
        'df[\'A\'] = df[\'A\']',
        set([]),
        set(['A'])
    ),
    # Test a string index with 1 offset
    (
        '=FUNC(Ab)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index with with 1 offset, positive
    (
        '=FUNC(Aa)',
        'A',
        'b',
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index with spaces
    (
        '=FUNC(Athis has spaces)',
        'A',
        'this has spaces',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['this has spaces', 'this also has spaces'])),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index that contains a column header, no shift
    (
        '=FUNC(AthisA)',
        'A',
        'thisA',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['thisA', 'also this'])),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index that contains a column header, shift
    (
        '=FUNC(AthisA)',
        'A',
        'also this',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['also this', 'thisA'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index that has prefixes
    (
        '=FUNC(Aaa)',
        'A',
        'b',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 3), index=pd.Index(['b', 'a', 'aa'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index that has prefixes with spaces
    (
        '=FUNC(Aa aa)',
        'A',
        'b',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 3), index=pd.Index(['b', 'a ', 'a aa'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime column has no fill value
    (
        '=FUNC(A2007-01-23 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame({'A': [pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')]}, index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a boolean column has no fill value
    (
        '=FUNC(A2007-01-23 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame({'A': [True]}, index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1))',
        set(['FUNC']),
        set(['A'])
    ),
]

INDEX_TEST_CASES_THAT_DONT_RECONSTRUCT_EXACTLY = [
    # Test references a column header and a range specifically
    (
        '=FUNC(A, A1)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a different number index
    (
        '=FUNC(A0.0, A2.0, A7)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 10), index=float64_index),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-2, fill_value=0), df[\'A\'].shift(-7, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references an index with a int dtype
    (
        '=FUNC(A0.0, A2.0)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.Index([0, 1, 2], dtype='int64')),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references to a column header that contains the index, and one that is unqualified
    (
        '=FUNC(HEADER00, HEADER0HEADER1)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['HEADER0', 'HEADER0HEADER1'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'HEADER0\'], df[\'HEADER0HEADER1\'])',
        set(['FUNC']),
        set(['HEADER0', 'HEADER0HEADER1'])
    ),
    # Test a string index with column header containing the index
    (
        '=FUNC(aaa)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'aaa\'])',
        set(['FUNC']),
        set(['aaa'])
    ),
    # Test a string index with column header containing the index, no shift
    (
        '=FUNC(aaaa)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'aaa\'])',
        set(['FUNC']),
        set(['aaa'])
    ),
    # Test a string index with column header containing the index, no shift
    (
        '=FUNC(aaab)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'aaa\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['aaa'])
    ),
    # Test a string index with column headers AND indexes that are prefixes, just detects column header
    (
        '=FUNC(aaaa)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['aaa', 'aaaa'], 2), index=pd.Index(['a', 'aa'])),
        'df[\'A\'] = FUNC(df[\'aaaa\'])',
        set(['FUNC']),
        set(['aaaa'])
    ),
]

HEADER_HEADER_RANGE_TEST_CASES = [
    # Simple header header reference
    (
        '=A:A',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'B\'] = df[[\'A\']]',
        set([]),
        set(['A'])
    ),
    # Header header reference in a formula
    (
        '=SUM(A:A)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'B\'] = SUM(df[[\'A\']])',
        set(['SUM']),
        set(['A'])
    ),
    # Multiple header header references
    (
        '=SUM(A:A, B:B)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'B\'] = SUM(df[[\'A\']], df[[\'B\']])',
        set(['SUM']),
        set(['A', "B"])
    ),
    # One range across two headers
    (
        '=SUM(A:B)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'B\'] = SUM(df.loc[:, \'A\':\'B\'])',
        set(['SUM']),
        set(['A', "B"])
    ),
    # One range across two headers, with as well as other specific cell reference
    (
        '=SUM(A:B, A0)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        "df[\'B\'] = SUM(df.loc[:, \'A\':\'B\'], df['A'])",
        set(['SUM']),
        set(['A', "B"])
    ),
    # One range across three headers, includes the middle one
    (
        '=SUM(A:C)',
        'D',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2), index=pd.RangeIndex(0, 2)),
        "df[\'D\'] = SUM(df.loc[:, \'A\':\'C\'])",
        set(['SUM']),
        set(["A", "B", "C"])
    ),
]

HEADER_INDEX_HEADER_INDEX_MATCHES = [
    # Simple header index header index reference
    (
        '=A0:A1',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        "df['B'] = RollingRange(df[['A']], 2, 0)",
        set([]),
        set(['A'])
    ),
    # Header index header index reference with offset
    (
        '=A1:A2',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.RangeIndex(0, 3)),
        "df['B'] = RollingRange(df[['A']], 2, 1)",
        set([]),
        set(['A'])
    ),
    # Header index header index reference with offset and two columns
    (
        '=A1:B2',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.RangeIndex(0, 3)),
        "df['B'] = RollingRange(df.loc[:, 'A':'B'], 2, 1)",
        set([]),
        set(['A', 'B'])
    ),
    # Header index header index reference with offset and two columns in function
    (
        '=SUM(A1:B2)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.RangeIndex(0, 3)),
        "df['B'] = SUM(RollingRange(df.loc[:, 'A':'B'], 2, 1))",
        set(['SUM']),
        set(['A', 'B'])
    ),
    # Header index header index reference with offset and two columns in function, as well as other specific cell reference and header ref
    (
        '=SUM(A1:B2, A0)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.RangeIndex(0, 3)),
        "df['B'] = SUM(RollingRange(df.loc[:, 'A':'B'], 2, 1), df['A'])",
        set(['SUM']),
        set(['A', 'B'])
    ),
    # Header index header index as well as header header
    (
        '=SUM(A1:B2, A:B)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.RangeIndex(0, 3)),
        "df['B'] = SUM(RollingRange(df.loc[:, 'A':'B'], 2, 1), df.loc[:, 'A':'B'])",
        set(['SUM']),
        set(['A', 'B'])
    ),
    # Header index header index reference with offset
    (
        '=A3:A7',
        'B',
        5,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 10), index=pd.RangeIndex(0, 10)),
        "df['B'] = RollingRange(df[['A']], 5, -2)",
        set([]),
        set(['A'])
    ),
    # Headers in incorrect direction in rolling range
    (
        '=SUM(B0:A0)',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.RangeIndex(0, 3)),
        "df['C'] = SUM(RollingRange(df.loc[:, 'A':'B'], 1, 0))",
        set(['SUM']),
        set(['B', 'A'])
    ),
    # One range across three headers, includes the middle one
    (
        '=SUM(C0:A0)',
        'D',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2), index=pd.RangeIndex(0, 2)),
        "df[\'D\'] = SUM(RollingRange(df.loc[:, 'A':'C'], 1, 0))",
        set(['SUM']),
        set(["A", "B", "C"])
    ),
]

"""
PARSE_TESTS contains a variety of tests that make sure
formula parsing is working properly; it is passed as 
a parameter into the test_parse test below.

Order of params is: formula, address, python_code, functions, columns

See documentation here: https://docs.pytest.org/en/latest/parametrize.html#parametrize-basics
"""
PARSE_TESTS = CONSTANT_TEST_CASES + OPERATOR_TEST_CASES + FUNCTION_TEST_CASES + INDEX_TEST_CASES + INDEX_TEST_CASES_THAT_DONT_RECONSTRUCT_EXACTLY + HEADER_HEADER_RANGE_TEST_CASES + HEADER_INDEX_HEADER_INDEX_MATCHES

@pytest.mark.parametrize("formula,column_header,formula_label,df,python_code,functions,columns", PARSE_TESTS)
def test_parse(formula, column_header, formula_label, df, python_code, functions, columns):
    code, funcs, cols, _ = parse_formula(formula, column_header, formula_label, {'type': FORMULA_ENTIRE_COLUMN_TYPE}, [df], ['df'], 0) 
    assert (code, funcs, cols) == \
        (
            python_code, 
            functions, 
            columns
        )


# Right now, VLOOKUP is the only formula that allows cross-sheet referencing. In the future, 
# other cross-sheet references in parser can be added here. 
VLOOKUP_TESTS = [
    (
        '=VLOOKUP(A0, df_2!C:D, 2)',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['C', 'D'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0,
        'df_1[\'B\'] = VLOOKUP(df_1[\'A\'], df_2.loc[:, \'C\':\'D\'], 2)',
        set(['VLOOKUP']),
        set(['A', 'D', 'C'])
    ),
    (
        '=VLOOKUP(A0, df_2!B:C, 2)',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['B', 'C'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0,
        'df_1[\'B\'] = VLOOKUP(df_1[\'A\'], df_2.loc[:, \'B\':\'C\'], 2)',
        set(['VLOOKUP']),
        set(['A', 'B', 'C'])
    ),
    # Test for setting the column headers to be the order that they appear in the dataframe after parsing
    (
        '=VLOOKUP(A0, df_2!C:B, 2)',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['B', 'C'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0,
        'df_1[\'B\'] = VLOOKUP(df_1[\'A\'], df_2.loc[:, \'B\':\'C\'], 2)',
        set(['VLOOKUP']),
        set(['A', 'B', 'C'])
    ),
    # Test for calling VLOOKUP inside another function
    (
        '=SUM(C0, VLOOKUP(A0, df_2!C:B, 2))',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B', 'C'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['B', 'C'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0,
        'df_1[\'B\'] = SUM(df_1[\'C\'], VLOOKUP(df_1[\'A\'], df_2.loc[:, \'B\':\'C\'], 2))',
        set(['VLOOKUP', 'SUM']),
        set(['A', 'B', 'C'])
    ),
    # Test for two calls to VLOOKUP
    (
        '=CONCAT(VLOOKUP(A0, df_2!C:D, 2), VLOOKUP(A0, df_2!C:E, 2))',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['C', 'D', 'E'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0,
        'df_1[\'B\'] = CONCAT(VLOOKUP(df_1[\'A\'], df_2.loc[:, \'C\':\'D\'], 2), VLOOKUP(df_1[\'A\'], df_2.loc[:, \'C\':\'E\'], 2))',
        set(['VLOOKUP', 'CONCAT']),
        set(['A', 'E', 'D', 'C'])
    )
]

@pytest.mark.parametrize("formula,column_header,formula_label,dfs,df_names,sheet_index,python_code,functions,columns", VLOOKUP_TESTS)
def test_parse_cross_sheet_formulas(formula, column_header, formula_label, dfs, df_names, sheet_index, python_code, functions, columns):
    code, funcs, cols, _ = parse_formula(formula, column_header, formula_label, {'type': FORMULA_ENTIRE_COLUMN_TYPE}, dfs, df_names, sheet_index) 
    assert (code, funcs, cols) == \
        (
            python_code, 
            functions, 
            columns
        )

POST_PD_1_2_VLOOKUP_TESTS = [
    # Test for cross-sheet reference with non-string column header
    (
        '=VLOOKUP(A0, df_2!1:C, 2)',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B', 'C'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(get_number_data_for_df([1, 'C'], 2))
        ],
        ['df_1', 'df_2'],
        0,
        'df_1[\'B\'] = VLOOKUP(df_1[\'A\'], df_2.loc[:, 1:\'C\'], 2)',
        set(['VLOOKUP']),
        set([1, 'A', 'C'])
    )
]

def test_specific_index_labels_header_header():
    formula = '=SUM(A:A)'
    column_header = 'B'
    formula_label = 0
    df = pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2))
    python_code = 'df.loc[[0], [\'B\']] = SUM(df[[\'A\']])'
    functions = set(['SUM'])
    columns = set(['A'])
    code, funcs, cols, _ = parse_formula(formula, column_header, formula_label, {'type': FORMULA_SPECIFIC_INDEX_LABELS_TYPE, 'index_labels': [0]}, [df], ['df'], 0) 
    assert (code, funcs, cols) == \
        (
            python_code, 
            functions, 
            columns
        )

def test_specific_index_labels_header_header_multiple_header_dependencies():
    formula = '=SUM(A:B)'
    column_header = 'C'
    formula_label = 0
    df = pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2), index=pd.RangeIndex(0, 2))
    python_code = 'df.loc[[0], [\'C\']] = SUM(df.loc[:, \'A\':\'B\'])'
    functions = set(['SUM'])
    columns = set(['A', 'B'])
    code, funcs, cols, _ = parse_formula(formula, column_header, formula_label, {'type': FORMULA_SPECIFIC_INDEX_LABELS_TYPE, 'index_labels': [0]}, [df], ['df'], 0) 
    assert (code, funcs, cols) == \
        (
            python_code, 
            functions, 
            columns
        )

@pytest.mark.parametrize("formula,column_header,formula_label,dfs,df_names,sheet_index,python_code,functions,columns", POST_PD_1_2_VLOOKUP_TESTS)
@pandas_post_1_2_only
def post_pandas_1_2_cross_sheet_tests(formula,column_header,formula_label,dfs,df_names,sheet_index,python_code,functions,columns):
    test_parse_cross_sheet_formulas(formula,column_header,formula_label,dfs,df_names,sheet_index,python_code,functions,columns)

# Right now, only {SHEET}{HEADER}{HEADER} is supported, so any other kind of cross-sheet reference should throw an error.
INVALID_VLOOKUP_TESTS = [
    (
        '=VLOOKUP(df_2!A0, df_2!C:D, 2)',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['C', 'D'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0
    ),
    (
        '=VLOOKUP(A0, df_2!C, 2)',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['C', 'D'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0
    ),
    (
        '=VLOOKUP(A0, C:D, df_1!)',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['C', 'D'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0
    ),
    (
        '=VLOOKUP(A0, df_2!C0:D0, 1)',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['C', 'D'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0
    ),
    (
        '=VLOOKUP(A0, df_2!C0, 1)',
        'B',
        0,
        [
            pd.DataFrame(
                get_number_data_for_df(['A', 'B'], 2),
                index=pd.RangeIndex(0, 2)
            ),
            pd.DataFrame(
                get_number_data_for_df(['C', 'D'], 2),
                index=pd.RangeIndex(0, 2)
            )
        ],
        ['df_1', 'df_2'],
        0
    )
]

@pytest.mark.parametrize("formula,column_header,formula_label,dfs,df_names,sheet_index", INVALID_VLOOKUP_TESTS)
def test_parse_invalid_cross_sheet_formulas(formula, column_header, formula_label, dfs, df_names, sheet_index):
    with pytest.raises(MitoError) as e_info:
        parse_formula(formula, column_header, formula_label, {'type': FORMULA_ENTIRE_COLUMN_TYPE}, dfs, df_names, sheet_index) 
    assert e_info.value.type_ == 'invalid_formula_error'
    assert e_info.value.to_fix == 'Cross-sheet references are only allowed for ranges of columns.'

PARSE_TEST_ERRORS = [
    ('=HLOOKUP(100, A)', 'B', 'invalid_formula_error', 'HLOOKUP'),
    ('=XLOOKUP(100, A)', 'B', 'invalid_formula_error', 'XLOOKUP'),
    ('=A <> 100', 'B', 'invalid_formula_error', '<>'),
    ('=SUM(A', 'B', 'invalid_formula_error', 'parentheses'),
    ('=A=B', 'B', 'invalid_formula_error', 'equality'),
    ('=IF(A=B, 1, 0)', 'B', 'invalid_formula_error', 'equality'),
    ('IF(A=B, 1, 0)', 'B', 'invalid_formula_error', 'equality'),
    ("A='='", 'B', 'invalid_formula_error', 'equality'),
    ("'=' = '='", 'B', 'invalid_formula_error', 'equality'),
    ("= = =", 'B', 'invalid_formula_error', 'equality'),
    ("A == A = A", 'B', 'invalid_formula_error', 'equality'),   
]
@pytest.mark.parametrize("formula, address, error_type, to_fix_substr", PARSE_TEST_ERRORS)
def test_parse_errors(formula, address, error_type, to_fix_substr):
    with pytest.raises(MitoError) as e_info:
        parse_formula(formula, address, 0, {'type': FORMULA_ENTIRE_COLUMN_TYPE}, [pd.DataFrame(get_number_data_for_df(['A', 'B'], 2))], ['df'], 0, 1)
    assert e_info.value.type_ == error_type
    if to_fix_substr is not None:
        assert to_fix_substr in e_info.value.to_fix


SAFE_CONTAINS_TESTS = [
    ('=A & "B"', '&', True),
    ('=A - +  "&"', '&', False),
    ('=A - +  \'&\'', '&', False),
]

@pytest.mark.parametrize('formula,substring,contains', SAFE_CONTAINS_TESTS)
def test_safe_contains(formula, substring, contains):
    assert safe_contains(formula, substring, ['A', 'B']) == contains


@pytest.mark.parametrize("formula,column_header,formula_label,df,python_code,functions,columns", INDEX_TEST_CASES + HEADER_HEADER_RANGE_TEST_CASES + HEADER_INDEX_HEADER_INDEX_MATCHES)
def test_get_frontend_formula_reconstucts_properly(formula,column_header,formula_label,df,python_code,functions,columns):
    frontend_formula = get_frontend_formula(formula, formula_label, [df], ['df'], 0)
    assert get_backend_formula_from_frontend_formula(frontend_formula, formula_label, df) == formula

@pytest.mark.parametrize("formula,column_header,formula_label,dfs,df_names,sheet_index,python_code,functions,columns", VLOOKUP_TESTS)
def test_get_cross_sheet_frontend_formula_reconstucts_properly(formula,column_header,formula_label,dfs,df_names,sheet_index,python_code,functions,columns):
    frontend_formula = get_frontend_formula(formula, formula_label, dfs, df_names, sheet_index)
    assert get_backend_formula_from_frontend_formula(frontend_formula, formula_label, dfs[sheet_index]) == formula


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import inspect
from functools import wraps
from typing import Any, Callable, Tuple

from mitosheet.errors import MitoError, make_function_error


def make_invalid_param_type_conversion_error(value: Any, target_type: str) -> MitoError:
    """
    When a user passes an input to a function that cannot be converted to the correct
    type, this error is thrown.
    """

    return MitoError(
        'invalid_param_type_conversion_error',
        'Param Type Conversion Error',
        f'The value {value} cannot be converted to {target_type}',
        error_modal=False
    )

def make_invalid_arg_error(sheet_function_name: str, arg_name: str, arg_index: int, type_name: str) -> MitoError:
    """
    When a user passes an input to a function that is the wrong type
    """

    arg_index_str = ''
    if arg_index == 0:
        arg_index_str = '1st'
    elif arg_index == 1:
        arg_index_str = '2nd'
    elif arg_index == 2:
        arg_index_str = '3rd'
    else:
        arg_index_str = f'{arg_index + 1}th'

    
    type_name_str = ''
    if type_name == 'DataFrame':
        type_name_str = 'an entire column reference'
    elif type_name == 'Series':
        type_name_str = 'a single cell reference'
    elif type_name == 'RollingRange':
        type_name_str = 'a range reference'
    else:
        type_name_str = f'a {type_name}.'

    arg_name_str = ''
    if arg_name != 'argv':
        arg_name_str = f' ({arg_name})'

    return MitoError(
        'invalid_arg_error',
        f'{sheet_function_name}: {arg_name} is Invalid',
        f'Error in {sheet_function_name}. The {arg_index_str} argument{arg_name_str} is invalid. It cannot be {type_name_str}.',
        error_modal=False
    )


def get_type_args(tp: Any) -> Tuple[Any, ...]:

    # If the type is just a single class, then we just return this type
    if inspect.isclass(tp):
        return (tp,)

    try:
        from typing import get_args
        return get_args(tp)
    except:
        # Handle older versions of Python without the get_args function
        if hasattr(tp, '__args__'):
            return tp.__args__
        elif hasattr(tp, '__orig_bases__'):
            return tp.__orig_bases__[0].__args__
        elif hasattr(tp, '__union_params__'): # handle Union types
            return tp.__union_params__
        else:
            return ()


def handle_sheet_function_errors(sheet_function: Callable) -> Callable:

    @wraps(sheet_function)
    def wrapped_sheet_function(*args):   

        parameters = inspect.signature(sheet_function).parameters

        # We ensure that all of the paramters match the type that is given to them
        for index, (parameter, arg) in enumerate(zip(parameters.values(), args)):
            if parameter.annotation == inspect.Parameter.empty:
                continue
            if not any(isinstance(arg, t) for t in get_type_args(parameter.annotation)):
                raise make_invalid_arg_error(sheet_function.__name__, parameter.name, index, type(arg).__name__)
        
        try:
            return sheet_function(*args)
        except MitoError:
            raise 
        except:
            raise make_function_error(sheet_function.__name__, error_modal=False)
    
    return wrapped_sheet_function




from copy import copy
from typing import List, Dict, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk



def optimize_code_chunks_combine_right(code_chunks_to_optimize: List[CodeChunk]) -> Tuple[bool, List[CodeChunk]]:
    """
    Given a list of code chunks, will attempt to optimize them right
    with a single pass.
    """

    code_chunks_to_optimize = copy(code_chunks_to_optimize)
    code_chunks_to_optimize.reverse() # Reverse the list so we can pop from the old front quickly

    code_chunks_list: List[CodeChunk] = []

    optimized = False
    while len(code_chunks_to_optimize) >= 2:
        later_code_chunk = code_chunks_to_optimize.pop()
        earlier_code_chunk = code_chunks_to_optimize.pop()

        combined_chunk = later_code_chunk.combine_right(earlier_code_chunk)

        if combined_chunk is not None:
            # If we can combine the two chunks into one, take that
            optimized = True
            code_chunks_to_optimize.append(combined_chunk)
        else:
            # If we cannot combine the two chunks, take the first chunk, 
            # and reset the second one to the new first chunk for the next loop
            code_chunks_list.append(later_code_chunk)
            code_chunks_to_optimize.append(earlier_code_chunk)            
    
    # Make sure we take the final item in the code chunks list, 
    # as it has nothing to combine_right with
    if len(code_chunks_to_optimize) == 1:
        code_chunks_list.append(code_chunks_to_optimize[0])

    return optimized, code_chunks_list

def optimize_code_chunks_combine_left(code_chunks_to_optimize: List[CodeChunk]) -> Tuple[bool, List[CodeChunk]]:
    """
    Given a list of code chunks, will attempt to optimize them left
    with a single pass.
    """

    code_chunks_to_optimize = copy(code_chunks_to_optimize)
    code_chunks_list: List[CodeChunk] = []

    optimized = False
    while len(code_chunks_to_optimize) >= 2:
        later_code_chunk = code_chunks_to_optimize.pop()
        earlier_code_chunk = code_chunks_to_optimize.pop()

        combined_chunk = later_code_chunk.combine_left(earlier_code_chunk)

        if combined_chunk is not None:
            # If we can combine the two chunks into one, take that
            optimized = True
            code_chunks_to_optimize.append(combined_chunk)
        else:
            # If we cannot combine the two chunks, take the first chunk, 
            # and reset the second one to the new first chunk for the next loop
            code_chunks_list.append(later_code_chunk)
            code_chunks_to_optimize.append(earlier_code_chunk)            
    
    # Make sure we take the final item in the code chunks list, 
    # as it has nothing to combine_left with
    if len(code_chunks_to_optimize) == 1:
        code_chunks_list.append(code_chunks_to_optimize[0])

    # Since we build the result backwards, we have to reverse it at the end
    code_chunks_list.reverse()

    return optimized, code_chunks_list


def reorder_code_chunks_for_more_optimization(all_code_chunks: List[CodeChunk]) -> Tuple[bool, List[CodeChunk]]:
    """
    Sometimes a user will perform actions in an order that is not super
    advantageous to code optimization. As such, we do our best to reorder
    code chunks in a way that:
    1. Absolutely is safe -- there should be no changes in behavior
    2. Results in a better code optimization

    The logic for when code chunks can be reordered can be found in 
    the can_be_reordered_with function defined on code chunks. 
    """
    if len(all_code_chunks) < 3:
        return False, all_code_chunks
    
    reordered = False
    final_code_chunks = [all_code_chunks[0], all_code_chunks[1]]
    for code_chunk in all_code_chunks[2:]:
        
        added = False
        for i in range(1, len(final_code_chunks)):
            one_back_code_chunk = final_code_chunks[-1 * i]
            two_back_code_chunk = final_code_chunks[-1 * i - 1]

            if one_back_code_chunk.can_be_reordered_with(code_chunk) and code_chunk.can_be_reordered_with(one_back_code_chunk):
                # Because this function is called after left and right optimization, we 
                # know that the one_back_code_chunk will not optimize with this one. So 
                # we check if this code chunk could be optimized with the two_back_code_chunk
                # and move it there if so. 
                optimized_right = two_back_code_chunk.combine_right(code_chunk) is not None
                optimized_left = code_chunk.combine_right(two_back_code_chunk) is not None

                if optimized_right or optimized_left:
                    final_code_chunks.insert(-1 * i, code_chunk)
                    reordered = True
                    added = True
                    break
            
            else:

                final_code_chunks.append(code_chunk)
                added = True
                break
        
        if not added:
            final_code_chunks.append(code_chunk)

    return reordered, final_code_chunks

def optimize_code_chunks(all_code_chunks: List[CodeChunk]) -> List[CodeChunk]:
    """
    Given a list of code chunks, will recursively attempt to optimize them 
    down to the smallest list of code chunks that have the same effects
    as the original list. 

    This is necessarily recursive, because of a situation like [A, A, B, B], 
    where A and B can be combined to a No-op. Thus, after one call, we end with 
    [A, B], and we need to recurse to finish the optimization.
    """
    # First, try and remove code chunks we know we can remove
    optimized_right, all_code_chunks = optimize_code_chunks_combine_right(all_code_chunks)
    optimized_left, all_code_chunks = optimize_code_chunks_combine_left(all_code_chunks)

    # Then, reorder the code chunks to be more optimal
    reordered, all_code_chunks = reorder_code_chunks_for_more_optimization(all_code_chunks)

    # As long as we optimized in this iteration, recurse as we might
    # need to optimize again
    if optimized_right or optimized_left or reordered:
        return optimize_code_chunks(all_code_chunks)

    return all_code_chunks


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Any, Dict, List, Union
from mitosheet.saved_analyses.save_utils import read_analysis
from mitosheet.types import StepsManagerType

def get_saved_analysis_code(params: Dict[str, Any], steps_manager: StepsManagerType) -> Union[List[str], str, None]:
    if steps_manager.analysis_to_replay is not None:
        analysis = read_analysis(steps_manager.analysis_to_replay)
        return analysis.get("code") if analysis is not None else None
    return None


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from collections import OrderedDict
import os
import pandas as pd
import pytest
from mitosheet.streamlit.v1 import spreadsheet, RunnableAnalysis
from mitosheet.tests.decorators import requires_streamlit

df1 = pd.DataFrame({'A': [123]})
df2 = pd.DataFrame({'A': [1234]})

SPREADSHEET_PARAMS = [
    (
        'Empty params',
        [], {},
        (
            {}, ''
        )
    ),
    (
        'A single dataframe',
        [df1], {},
        (
            OrderedDict({'df1': df1}), ''
        )
    ),
    (
        'Multiple dataframes',
        [df1, df2, df1], {},
        (
            OrderedDict({'df1': df1, 'df2': df2, 'df3': df1}), ''
        )
    ),
    (
        'return type of code',
        [df1], {'return_type': 'default_list'},
        (
            [df1], ''
        )
    ),
    (
        'return type of code',
        [df1], {'return_type': 'code'},
        (
            ''
        )
    ),
    (
        'return type of dfs',
        [df1], {'return_type': 'dfs_dict'},
        (
            OrderedDict({'df1': df1})
        )
    ),
    (
        'return type of dfs_list',
        [df1], {'return_type': 'dfs_list'},
        (
            [df1]
        )
    ),
    (
        'return type of analysis',
        [df1], {'return_type': 'analysis'},
        (
            RunnableAnalysis('', None, '', [], 0)
        )
    )
]

@pytest.mark.parametrize('test, args, kwargs, expected', SPREADSHEET_PARAMS)
@requires_streamlit
def test_creates_spreadsheet(test, args, kwargs, expected):

    result = spreadsheet(*args, **kwargs)

    if isinstance(result, tuple):
        for v1, v2 in zip(result, expected):
            if isinstance(v1, OrderedDict):
                assert len(v1) == len(v2)
                for k in v1:
                    assert v1[k].equals(v2[k])
            elif isinstance(v1, list):
                for e1, e2 in zip(v1, v2):
                    assert e1.equals(e2)
            else:
                assert v1 == v2
    elif isinstance(result, list):
        for e1, e2 in zip(result, expected):
            assert e1.equals(e2)
    elif isinstance(result, OrderedDict):
        assert len(result) == len(expected)
        for k in result:
            assert result[k].equals(expected[k])
    elif isinstance(result, RunnableAnalysis):
        assert isinstance(expected, RunnableAnalysis)
    else:
        assert result == expected

@requires_streamlit
def test_return_type_function():
    f = spreadsheet(df1, code_options={'as_function': True, 'call_function': False, 'function_name': 'test', 'function_params': {}}, return_type='function')
    assert callable(f)

@requires_streamlit
def test_return_type_function_invalid_code_options():
    with pytest.raises(ValueError):
        spreadsheet(df1, return_type='function')
    with pytest.raises(ValueError):
        spreadsheet(df1, code_options={'as_function': False, 'call_function': False, 'function_name': 'test', 'function_params': {}}, return_type='function')
    with pytest.raises(ValueError):
        spreadsheet(df1, code_options={'as_function': True, 'call_function': True, 'function_name': 'test', 'function_params': {}}, return_type='function')

@requires_streamlit   
def test_spreadsheet_with_column_definitions():
    f = spreadsheet(
        df1, 
        column_definitions=[
            [
                {
                    'columns': ['A'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'greater_than_or_equal', 'value': 5}], 
                        'font_color': '#c30010', 
                        'background_color': '#ffcbd1' 
                    }] 
                },
                {
                    'columns': ['A'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'less', 'value': 2}], 
                        'font_color': '#f30010', 
                        'background_color': '#ddcbd1' 
                    }] 
                }
            ]
        ], 
        code_options={'as_function': True, 'call_function': False, 'function_name': 'test', 'function_params': {}}, 
        return_type='function'
    )
    assert callable(f)

@requires_streamlit   
def test_spreadsheet_with_column_definitions_only_one_color():
    f = spreadsheet(
        df1, 
        column_definitions=[
            [
                {
                    'columns': ['A'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'greater_than_or_equal', 'value': 5}], 
                        'font_color': '#c30010', 
                    }] 
                },
                {
                    'columns': ['A'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'less', 'value': 2}], 
                        'background_color': '#ddcbd1' 
                    }] 
                }
            ]
        ], 
        code_options={'as_function': True, 'call_function': False, 'function_name': 'test', 'function_params': {}}, 
        return_type='function'
    )
    assert callable(f)

@requires_streamlit   
def test_spreadsheet_with_default_apply_formula_to_column():
    f = spreadsheet(
        df1, 
        default_editing_mode='cell',
        code_options={'as_function': True, 'call_function': False, 'function_name': 'test', 'function_params': {}}, 
        return_type='function'
    )
    assert callable(f)


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mitosheet.public.v2 import (
    sheet_functions, utils, flatten_column_header, deduplicate_column_headers,
    to_int_series, to_boolean_series, to_float_series, to_timedelta_series, get_datetime_format)

from mitosheet.public.v3.rolling_range import RollingRange
from mitosheet.public.v3.formatting import add_formatting_to_excel_sheet
from mitosheet.public.v3.sheet_functions import FUNCTIONS
from mitosheet.public.v3.sheet_functions import *
from mitosheet.public.v3.types.bool import cast_string_to_bool

from mitosheet.public.v1 import register_analysis
from mitosheet.public.v2.excel_utils import get_table_range_from_upper_left_corner_value, get_read_excel_params_from_range, get_table_range, convert_csv_file_to_xlsx_file
import pandas as pd



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Useful decorators for tests. Specifically, we often only want
to run specific tests on specific versions of pandas or Python
"""

import os
import pytest
import pandas as pd
import sys
from mitosheet.ai.ai_utils import is_open_ai_credentials_available

from mitosheet.utils import is_flask_installed, is_prev_version, is_snowflake_connector_python_installed, is_snowflake_credentials_available, is_streamlit_installed, is_dash_installed

pandas_pre_1_only = pytest.mark.skipif(
    not pd.__version__.startswith('0.'), 
    reason='This test only runs on earlier versions of Pandas. API inconsistencies make it fail on earlier versions'
)

pandas_pre_2_only = pytest.mark.skipif(
    pd.__version__.startswith('2.'), 
    reason='This test only runs on earlier versions of Pandas. API inconsistencies make it fail on earlier versions'
)

pandas_post_0_24_0_only = pytest.mark.skipif(
    is_prev_version(pd.__version__, '0.24.0'), 
    reason='This test only runs on later versions of Pandas. API inconsistencies make it fail on earlier versions'
)

pandas_post_1_only = pytest.mark.skipif(
    pd.__version__.startswith('0.'), 
    reason='This test only runs on later versions of Pandas. API inconsistencies make it fail on earlier versions'
)

pandas_post_1_2_only = pytest.mark.skipif(
    is_prev_version(pd.__version__, '1.2.0'), 
    reason='This test only runs on later versions of Pandas. API inconsistencies make it fail on earlier versions'
)

pandas_post_1_4_only = pytest.mark.skipif(
    is_prev_version(pd.__version__, '1.4.0'), 
    reason='This test only runs on later versions of Pandas. API inconsistencies make it fail on earlier versions'
)

pandas_post_1_5_only = pytest.mark.skipif(
    is_prev_version(pd.__version__, '1.5.0'), 
    reason='This test only runs on later versions of Pandas. API inconsistencies make it fail on earlier versions'
)

pandas_pre_1_2_only = pytest.mark.skipif(
    not is_prev_version(pd.__version__, '1.2.0'), 
    reason='This test only runs on later versions of Pandas. API inconsistencies make it fail on earlier versions'
)

python_post_3_6_only = pytest.mark.skipif(
    sys.version_info.minor <= 6, 
    reason="requires 3.7 or greater"
)

requires_snowflake_dependencies_and_credentials = pytest.mark.skipif(
    not is_snowflake_connector_python_installed() or not is_snowflake_credentials_available(),
    reason='requires snowflake_connector_python package and snowflake credentials'
)

requires_streamlit = pytest.mark.skipif(
    not is_streamlit_installed(),
    reason='requires streamlit to be installed'
)

requires_flask = pytest.mark.skipif(
    not is_flask_installed(),
    reason='requires flask to be installed'
)

requires_dash = pytest.mark.skipif(
    not is_dash_installed(),
    reason='requires dash to be installed'
)

requires_open_ai_credentials = pytest.mark.skipif(
    not is_open_ai_credentials_available(),
    reason='Requires a set OPENAI_API_KEY'
)

only_on_github_actions = pytest.mark.skipif(
    'GITHUB_ACTIONS' not in os.environ,
    reason='This test only runs on GitHub Actions'
)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.types import ColumnID
from mitosheet.state import State

from mitosheet.errors import make_invalid_column_type_change_error
from mitosheet.is_type_utils import ( is_bool_dtype,
                                                    is_datetime_dtype,
                                                    is_float_dtype,
                                                    is_int_dtype,
                                                    is_string_dtype,
                                                    is_timedelta_dtype)
from mitosheet.transpiler.transpile_utils import get_column_header_as_transpiled_code, get_param_dict_as_code


def get_conversion_code(state: State, sheet_index: int, column_id: ColumnID, old_dtype: str, new_dtype: str, to_datetime_params_map: Optional[Dict[ColumnID, Dict[str, Any]]], public_interface_version: int) -> Optional[str]:
    
    column_header = state.column_ids.get_column_header_by_id(sheet_index, column_id)
    transpiled_column_header = get_column_header_as_transpiled_code(column_header)
    df_name = state.df_names[sheet_index]

    if is_bool_dtype(old_dtype):
        if is_bool_dtype(new_dtype):
            return None
        elif is_int_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].astype(\'int\')'
        elif is_float_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].astype(\'float\')'
        elif is_string_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].astype(\'str\')'
        elif is_datetime_dtype(new_dtype):
            raise make_invalid_column_type_change_error(
                column_header,
                old_dtype,
                new_dtype
            )
        elif is_timedelta_dtype(new_dtype):
            raise make_invalid_column_type_change_error(
                column_header,
                old_dtype,
                new_dtype
            )
    elif is_int_dtype(old_dtype):
        if is_bool_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].fillna(False).astype(\'bool\')'
        elif is_int_dtype(new_dtype):
            return None
        elif is_float_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].astype(\'float\')'
        elif is_string_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].astype(\'str\')'
        elif is_datetime_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = pd.to_datetime({df_name}[{transpiled_column_header}], unit=\'s\', errors=\'coerce\')'
        elif is_timedelta_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = pd.to_timedelta({df_name}[{transpiled_column_header}], unit=\'s\', errors=\'coerce\')'
    elif is_float_dtype(old_dtype):
        if is_bool_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].fillna(False).astype(\'bool\')'
        elif is_int_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].fillna(0).astype(\'int\')'
        elif is_float_dtype(new_dtype):
            return None
        elif is_string_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].astype(\'str\')'
        elif is_datetime_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = pd.to_datetime({df_name}[{transpiled_column_header}], unit=\'s\', errors=\'coerce\')'
        elif is_timedelta_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = pd.to_timedelta({df_name}[{transpiled_column_header}], unit=\'s\', errors=\'coerce\')'
    elif is_string_dtype(old_dtype):
        if is_bool_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = to_boolean_series({df_name}[{transpiled_column_header}])'
        elif is_int_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = to_int_series({df_name}[{transpiled_column_header}])'
        elif is_float_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = to_float_series({df_name}[{transpiled_column_header}])'
        elif is_string_dtype(new_dtype):
            return None
        elif is_datetime_dtype(new_dtype):
            if to_datetime_params_map is not None:
                to_datetime_params = to_datetime_params_map[column_id]
            else:
                to_datetime_params = {}

            datetime_params_string = get_param_dict_as_code(to_datetime_params, as_single_line=True)

            if datetime_params_string:
                return f'{df_name}[{transpiled_column_header}] = pd.to_datetime({df_name}[{transpiled_column_header}], {datetime_params_string}, errors=\'coerce\')'
            else:
                return f'{df_name}[{transpiled_column_header}] = pd.to_datetime({df_name}[{transpiled_column_header}], errors=\'coerce\')'
        elif is_timedelta_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = pd.to_timedelta({df_name}[{transpiled_column_header}], errors=\'coerce\')'
    elif is_datetime_dtype(old_dtype):
        if is_bool_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = ~{df_name}[{transpiled_column_header}].isnull()'
        elif is_int_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].astype(\'int\') / 10**9'
        elif is_float_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].astype(\'int\').astype(\'float\') / 10**9'
        elif is_string_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].dt.strftime(\'%Y-%m-%d %X\')'
        elif is_datetime_dtype(new_dtype):
            return None
        elif is_timedelta_dtype(new_dtype):
            raise make_invalid_column_type_change_error(
                column_header,
                old_dtype,
                new_dtype
            )
    elif is_timedelta_dtype(old_dtype):
        if is_bool_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = ~{df_name}[{transpiled_column_header}].isnull()'
        elif is_int_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].dt.total_seconds().astype(\'int\')'
        elif is_float_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].dt.total_seconds()'
        elif is_string_dtype(new_dtype):
            return f'{df_name}[{transpiled_column_header}] = {df_name}[{transpiled_column_header}].astype(\'str\')'
        elif is_datetime_dtype(new_dtype):
            raise make_invalid_column_type_change_error(
                column_header,
                old_dtype,
                new_dtype
            )
        elif is_timedelta_dtype(new_dtype):
            return None

    return None
    



class ChangeColumnDtypeCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, column_ids: List[ColumnID], old_dtypes: Dict[ColumnID, str], new_dtype: str, changed_column_ids: List[ColumnID], to_datetime_params_map: Optional[Dict[ColumnID, Dict[str, Any]]], public_interface_version: int):
        super().__init__(prev_state)
        self.sheet_index = sheet_index
        self.column_ids = column_ids
        self.old_dtypes = old_dtypes
        self.new_dtype = new_dtype
        self.changed_column_ids = changed_column_ids
        self.to_datetime_params_map = to_datetime_params_map
        self.public_interface_version = public_interface_version

        self.df_name = self.prev_state.df_names[self.sheet_index]
        self.column_headers = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index, self.changed_column_ids)

    def get_display_name(self) -> str:
        return 'Changed dtype'
    
    def get_description_comment(self) -> str:
        return f'Changed {", ".join([str(ch) for ch in self.column_headers])} to dtype {self.new_dtype}'

    def get_code(self) -> Tuple[List[str], List[str]]:

        # Note: we can't actually group all the headers together in one conversion, and not every dtype can be converted
        # to the target dtype in the same way. Even if they have the same old_dtype, this still might not work in the case
        # of converting to datetimes. It ends up being really hard to actually group these together (>50+ lines of relatively
        # dense and complex code), so we avoid it for now.

        code = []
        for column_id in self.column_ids:
            old_dtype = self.old_dtypes[column_id]

            conversion_code = get_conversion_code(self.prev_state, self.sheet_index, column_id, old_dtype, self.new_dtype, self.to_datetime_params_map, self.public_interface_version)
            if conversion_code is not None:
                code.append(conversion_code)
        
        # If we have pandas included, then add pandas as an import
        return code, ['import pandas as pd'] if any('pd.to_datetime' in line for line in code) else []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Utilities to help with type functions
"""

from packaging.version import Version
from typing import Any, Dict, List, Optional, Tuple, Union
import pandas as pd
import numpy as np
import datetime

from mitosheet.public.v1.sheet_functions.sheet_function_utils import is_series_of_constant


def get_nan_indexes_metadata(*argv: pd.Series) -> Tuple[pd.Index, pd.Index]: 
    """
    Given a list of series, this function returns data that is helpful
    in figuring out which of the rows of these series have a NaN in them.

    The data returned allows you to remove the nan values from the series,
    so that they can then be easily used by sheet functions, but also to
    easily add these NaN values back to the series in the correct location.

    It does so by returning a tuple of the original_index, non_nan_index 
    """
    nan_index_set = set()
    non_nan_index_set = set()
    original_index = None

    for arg in argv:
        if isinstance(arg, pd.Series): 
            nan_index_set.update(arg[arg.isnull()].index)
            non_nan_index_set.update(arg[~arg.isnull()].index)

            # Update the original indexes to the first element, if we haven't
            # selected one yet
            if original_index is None:
                original_index = arg.index
            elif not is_series_of_constant(arg):
                original_index = arg.index

    # Remove any index that was nan at any point
    non_nan_index_set.difference_update(nan_index_set)
    non_nan_index = pd.Index(list(non_nan_index_set))

    return original_index, non_nan_index 

def put_nan_indexes_back(series: pd.Series, original_index: pd.Index) -> pd.Series:
    """
    This function takes a series, as well as a set of the indexes
    that are are NaN, and inserts these indexes back into the series
    as NaN values.
    """
    return series.reindex(original_index)


def get_to_datetime_params(string_series: pd.Series) -> Dict[str, Any]:

    # If the series is already full of datetime objects, then we don't need to 
    # guess a datetime format. 
    if is_series_full_of_datetimes(string_series):
        return {}


    detected_format = get_datetime_format(string_series)

    # If we detect a format, we return that. This works for all pandas versions
    if detected_format is not None:
        return {
            'format': detected_format
        }
    

    # If pandas < 2.0, we can use infer_datetime_format
    if Version(pd.__version__) < Version('2.0'):
        return {
            'infer_datetime_format': True
        }
    
    # Otherwise, we mark the format as mixed
    return {
        'format': 'mixed'
    }

def is_series_full_of_datetimes(string_series: pd.Series) -> bool:
    return all(isinstance(x, datetime.datetime) for x in string_series)


def get_datetime_format(string_series: pd.Series) -> Optional[str]:
    """
    Given a series of datetime strings, guesses the most likely date format.
    """
    # Import log function here to avoid circular import
    from mitosheet.telemetry.telemetry_utils import log
    
    # Filter to only the strings since that is all we're convertin
    string_series = string_series[string_series.apply(lambda x: isinstance(x, str))]

    # If we can convert all non null inputs, then we assume we guessed correctly
    non_null_inputs = string_series[~string_series.isna()]

    # If we're on a older verison of pandas, we just check if infer_datetime_format=True works
    if Version(pd.__version__) < Version('2.0'):
        converted = pd.to_datetime(non_null_inputs, errors='coerce', infer_datetime_format=True)
        if converted.isna().sum() == 0:
            return None

    # TODO: Add the most popular formats to here and check them first before 
    # trying all of the formats below for performance.
    sample_string_datetime = string_series[string_series.first_valid_index()]
    FORMATS = [
        '%m{s}%d{s}%Y', 
        '%d{s}%m{s}%Y', 
        '%Y{s}%m{s}%d', 
        '%Y{s}%d{s}%m', 
        '%d{s}%m{s}%Y  %H:%M:%S', 
        '%m{s}%d{s}%Y %H:%M:%S', 
        '%Y{s}%m{s}%d  %H:%M:%S',
        '%Y{s}%d{s}%m  %H:%M:%S', 

    ]
    SEPERATORS = ['/', '-', '.', ',', ':', ' ', '']

    for seperator in SEPERATORS:
        if seperator in sample_string_datetime:
            for _format in FORMATS:
                format = _format.format(s=seperator)
                # If we fail to convert to a specific format,keep trying
                try:
                    if test_datetime_format(non_null_inputs, format):
                        return format
                except:
                    pass

    log('unable_to_determine_datetime_format_on_cast')
    return None


def test_datetime_format(series: pd.Series, _format: str) -> bool:
    converted = pd.to_datetime(series, errors='coerce', format=_format)
    return converted.isna().sum() == 0
        

def get_million_identifier_in_string(string: str) -> Union[str, None]:
    """
    Given a string, returns the million identifier in it. 
    Returns '' if none exist. 
    """
    million_identifiers = ["Million", 'Mil', 'M', 'million', 'mil', 'm']
    # So that we return the biggest matching element
    million_identifiers = list(sorted(million_identifiers, key = len, reverse=True))

    for identifier in million_identifiers:
        if identifier in string:
            return identifier

    return None


def get_billion_identifier_in_string(string: str) -> Union[str, None]:
    """
    Given a string, returns the billion identifier in it. 
    Returns '' if none exist. 
    """

    billion_identifiers = ["Billion", 'Bil', 'B', 'billion', 'bil', 'b']
    # So that we return the biggest matching element
    billion_identifiers = list(sorted(billion_identifiers, key = len, reverse=True))

    for identifier in billion_identifiers:
        if identifier in string:
            return identifier

    return None



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of The Mito Enterprise license.

from typing import Any, Dict, List, Optional, Set

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.one_hot_encoding_code_chunk import \
    OneHotEncodingCodeChunk
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param


class OneHotEncodingStepPerformer(StepPerformer):
    """
    Allows you to do a one hot encoding of a column.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'one_hot_encoding'


    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            OneHotEncodingCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'), 
                get_param(params, 'column_id'), 
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Changes a specific field in the user.json file
"""

from typing import Any
from mitosheet.telemetry.telemetry_utils import identify
from mitosheet.types import StepsManagerType
from mitosheet.user import set_user_field


SET_USER_FIELD_EVENT = 'set_user_field_update'
SET_USER_FIELD_PARAMS = ['field', 'value']


def execute_set_user_field_update(steps_manager: StepsManagerType, field: str, value: Any) -> None:
    """
    The function responsible for setting specific field values
    for the user.json file, and then reidentifying the user
    """
    # Set the user_email in user.json
    set_user_field(field, value)

    # And we also reidentify after this was set, in case anything
    # about the users identity changed
    identify()

SET_USER_FIELD_UPDATE = {
    'event_type': SET_USER_FIELD_EVENT,
    'params': SET_USER_FIELD_PARAMS,
    'execute': execute_set_user_field_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains exports for all functions that can be used in the mitosheet
"""
from mitosheet.public.v1.sheet_functions.number_functions import *
from mitosheet.public.v1.sheet_functions.string_functions import *
from mitosheet.public.v1.sheet_functions.date_functions import *
from mitosheet.public.v1.sheet_functions.control_functions import *
from mitosheet.public.v1.sheet_functions.misc_functions import *

FUNCTIONS = dict(NUMBER_FUNCTIONS, **STRING_FUNCTIONS, **DATE_FUNCTIONS, **CONTROL_FUNCTIONS, **MISC_FUNCTIONS)

# Overwrite __all__ so when you run from mitosheet.sheet_functions import *, it just imports the functions themselves!
__all__ = [
    func for func in FUNCTIONS.keys()
]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
Helpers for accessing the user.json file
"""
import os
import json
from typing import Any, Dict, Optional
from mitosheet.save_paths import MITO_FOLDER

# The path of the user.json file
USER_JSON_PATH = os.path.join(MITO_FOLDER, 'user.json')

def get_user_json_object() -> Optional[Dict[str, Any]]:
    """
    Gets the entire user json object
    """
    try:
        with open(USER_JSON_PATH) as f:
            return json.load(f)
    except: 
        return None

def get_user_field(field: str) -> Optional[Any]:
    """
    Returns the value stored at field in the user.json file,
    but may read a different file if it passed
    """
    try:
        with open(USER_JSON_PATH) as f:
            return json.load(f)[field]
    except: 
        return None

def set_user_json_object(user_json_object: Dict[str, Any]) -> None:
    """
    Updates the value of a specific feild in user.json
    """
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps(user_json_object))

def set_user_field(field: str, value: Any) -> None:
    """
    Updates the value of a specific feild in user.json
    """
    with open(USER_JSON_PATH, 'r') as user_file_old:
        old_user_json = json.load(user_file_old)
        old_user_json[field] = value
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(old_user_json))



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mitosheet.startup.startup_utils import (
    create_startup_file, remove_startup_file 
)